from io import BytesIO
from PIL import Image
from typing import Any
from pydantic import BaseModel
from cfclient.core import HttpClient
from cfclient.core import ClientSession
from cflearn.parameters import OPT


def get_url(endpoint: str) -> str:
    return f"{OPT.host}{endpoint}"


class get_http_session:
    def __init__(self):
        self._http_client = None

    async def __aenter__(self) -> ClientSession:
        self._http_client = HttpClient()
        self._http_client.start()
        return self._http_client.session

    async def __aexit__(self, *args: Any) -> None:
        await self._http_client.stop()
        self._http_client = None


async def get_image_res(url: str, model: BaseModel) -> Image.Image:
    async with get_http_session() as session:
        async with session.post(url, json=model.dict()) as response:
            return Image.open(BytesIO(await response.read()))


__all__ = [
    "get_url",
    "get_http_session",
    "get_image_res",
]



import cv2
import math

import numpy as np

from typing import List


def resize_image(input_image: np.ndarray, resolution: int) -> np.ndarray:
    H, W, C = input_image.shape
    H = float(H)
    W = float(W)
    k = float(resolution) / min(H, W)
    H *= k
    W *= k
    H = int(np.round(H / 64.0)) * 64
    W = int(np.round(W / 64.0)) * 64
    img = cv2.resize(
        input_image,
        (W, H),
        interpolation=cv2.INTER_LANCZOS4 if k > 1 else cv2.INTER_AREA,
    )
    return img


def to_canvas(results: List[np.ndarray], *, padding: int = 0) -> np.ndarray:
    num_results = len(results)
    if num_results == 1:
        return results[0]
    num_col = math.ceil(math.sqrt(num_results))
    num_row = round(num_results / num_col)
    if num_row * num_col < num_results:
        num_row += 1
    h, w = results[0].shape[:2]
    canvas_w = num_col * w + (num_col - 1) * padding
    canvas_h = num_row * h + (num_row - 1) * padding
    canvas = np.full([canvas_h, canvas_w, 3], 255, np.uint8)
    for i, out in enumerate(results):
        ih, iw = out.shape[:2]
        if h != ih:
            raise ValueError(f"`h` mismatch: {ih} != {h}")
        if w != iw:
            raise ValueError(f"`w` mismatchh: {iw} != {w}")
        ix = i % num_col
        iy = i // num_col
        ix = ix * w + ix * padding
        iy = iy * h + iy * padding
        canvas[iy : iy + h, ix : ix + w] = out
    return canvas



import time
import torch

from PIL import Image
from typing import Any
from cftool.cv import to_rgb
from cftool.cv import restrict_wh
from cflearn.api.cv.third_party.blip import BLIPAPI

from .common import IAlgorithm
from .common import TextModel
from .common import ImageModel
from .common import MaxWHModel
from .parameters import init_to_cpu
from .parameters import auto_lazy_load
from .parameters import need_change_device


img2txt_caption_endpoint = "/img2txt/caption"


class Img2TxtModel(MaxWHModel, ImageModel):
    pass


@IAlgorithm.auto_register()
class Img2TxtCaption(IAlgorithm):
    model_class = Img2TxtModel
    response_model_class = TextModel

    endpoint = img2txt_caption_endpoint

    def initialize(self) -> None:
        self.lazy = auto_lazy_load()
        print(f"> init blip{' (lazy)' if self.lazy else ''}")
        self.m = BLIPAPI("cpu" if init_to_cpu() or self.lazy else "cuda:0")

    async def run(self, data: Img2TxtModel, *args: Any) -> TextModel:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        w, h = image.size
        w, h = restrict_wh(w, h, data.max_wh)
        image = image.resize((w, h), resample=Image.LANCZOS)
        t2 = time.time()
        if need_change_device() or self.lazy:
            self.m.to("cuda:0")
        t3 = time.time()
        caption = self.m.caption(to_rgb(image))
        t4 = time.time()
        if need_change_device() or self.lazy:
            self.m.to("cpu")
            torch.cuda.empty_cache()
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "get_model": t3 - t2,
                "inference": t4 - t3,
                "cleanup": time.time() - t4,
            }
        )
        return TextModel(text=caption)


__all__ = [
    "img2txt_caption_endpoint",
    "Img2TxtCaption",
]



# `direct` sdk is used to access to the server launched by `cfcreator serve`

from cfcreator import *
from PIL.Image import Image

from .utils import *


async def txt2img(model: Txt2ImgSDModel) -> Image:
    return await get_image_res(get_url(txt2img_sd_endpoint), model)


__all__ = [
    "txt2img",
]
